﻿namespace Tracer

module Lighting =
    type Point = Point.Point
    type Color = Color.Color
    type Ray = Ray.Ray
    type Shape = Shape.IShape
    type Vector = Vector.Vector

    (*get shadow rays from a point to a list of points*)
    let getShadowRays (pointlist: Point list) (origin:Point)  : Ray list = 
        let getARay lightpos o=
            let vectorBtwn = Point.distance o lightpos                       //Vector between hit position and light position
            let vectorBtwnInv = Vector.mkVector (1.0 / (Vector.getX vectorBtwn)) (1.0 / (Vector.getY vectorBtwn)) (1.0 / (Vector.getZ vectorBtwn))
            let distance = Vector.magnitude vectorBtwn               
            Ray.mkShadow o (Vector.normalise vectorBtwn) vectorBtwnInv distance //Shadow ray which is a ray from rayHitPosition to light source   
        List.map (fun p -> getARay p origin) pointlist
    
    (*functions to weigh the light by (eg. cosine similarity)*)
    let cosweight sr hit =
        let n = snd hit
        let c = Vector.cosBtwn n (Ray.getDirection sr) //cosine similarity
        c
    let distweight sr hit =
        let dist = Ray.getDistance sr
        1.0/dist**2.0 //inverse square law
    let noweight sr hit = 1.0   //a dummy that doesn't affect the weight. the alternative to having this is to make the weight list parameter for getLighting be an Option, and then just use None instead of this function. But the relative advantages and disadvantages of each option seem very minor.

    (*determine the lighting on a surface*)
    let Lighting color intensity weights (rays:Ray list) (shapes:Shape list) hit =  
        let lightPer sr= //light (color) per shadow ray
            let lr, lg, lb = Color.getRGB color //the light's r g b
            let wr, wg, wb = //weighted r g b
                let weight = 
                    let weight1 = List.fold (fun acc w -> acc*(w sr hit)) 1.0 weights //applies each weight.
                    weight1 * intensity * (1.0/(float)rays.Length)  //divide by amount of rays, so if half of the rays hit the overall then the added light is half of what it'd be if they all hit.
                weight * lr, weight * lg, weight * lb
            Color.mkColor wr wg wb

        let lightforray ray =
            let ifdoeshit option = //helper function
                match option with
                | Some option -> true
                | None -> false
            let rayhitsanything = List.tryFind (fun (s:Shape) -> ifdoeshit (s.Hit ray)) shapes //check if the ray is blocked
            match rayhitsanything with
            | Some h -> None
            | None -> Some (lightPer ray)   //if it's not blocked, get the light for the ray
        let colors = List.choose(fun r->lightforray r) rays
        if List.isEmpty colors then None else Some (Color.addColors colors)
  
  module Lights =
    open Color
    open Lighting

    (*Ambient light and the application thereof*)
    type AmbientLight = | AL of Color * float
    let mkAmbientLight color intensity = AL(color, intensity)
    let getColor (AL(c,_)) = c
    let getIntensity (AL(_,i)) = i
    let applyAmbient (ambientLight : AmbientLight) (color: Color) : Color =
        let Ia = getIntensity ambientLight          //Intensity of the ambient light
        let ambientColor = getColor ambientLight
        let ar, ag, ab = getRGB ambientColor
        let cr, cg, cb = getRGB color
        Color.Clr(cr + (ar * Ia), cg + (ag * Ia) , cb + (ab * Ia))

    (*ILight and construction of different kinds of light*)
    type ILight =
        abstract member getShadowRays: (Point -> (Ray list))
        abstract member getLighting: ((Ray->float*Vector.Vector -> float)list -> Ray list -> Shape list -> float*Vector -> Option<Color>) //weight list in the front so you can choose to use it when constructing a light
        abstract member getDoesHit: (Ray->bool)

    let mkPointLight (position: Point) (color:Color) (intensity:float)=
        {
         new ILight with 
         member this.getShadowRays = getShadowRays [position]
         member this.getLighting = Lighting color intensity//remaining: weight list, ray list, shape list
         member this.getDoesHit = (fun r -> false)
         }
    let mkSquareLight (samplefunc: Point -> Point -> Point -> Point list) (color:Color) (intensity:float) p1 p2 p3=
        {
         new ILight with 
         member this.getShadowRays = getShadowRays (samplefunc p1 p2 p3)
         member this.getLighting = Lighting color intensity
         member this.getDoesHit = (fun r -> match Shape.hitSide p1 p2 p3 r with
                                            |Some a -> true
                                            |None -> false        )
        }