﻿namespace Tracer
module Trace = 
    open Vector
    open Point
    open Ray
    open Camera
    open Scene
    open Color
    open System.Drawing
    open Lighting
    open Lights
    open Shape
    open Texture
    open Render
    open Scatter

    (*colors*)
    let white = fromSysColor Color.White
    let red = fromSysColor Color.Red

    (*texture functions*)
    let noise p = tomonochrome (valuenoise p)
    let fbm p = tomonochrome (fBm 5 p 2.0 0.5 false)

    let marble p = tomonochrome (marbletex p)
    let colormarble p = tocolor (marbletex p) defaultramp

    let sandstone p = tomonochrome (sandstonetex p)
    let colorsandstone p = tocolor (sandstonetex p) defaultramp

    let imgsandstone p = toimgcolor (sandstonetex p) (new Bitmap("texsandstone.png"))
    let imgrainbow p = toimgcolor (marbletex p) (new Bitmap("texrainbow.png"))
    let imgpinkblue p = toimgcolor (marbletex p) (new Bitmap("texpinkblue.png"))

    (*sample functions for lights, though not yet given a location in space.)*)
    let randomsample = RandomSampleSquare 16
    let evensample = EvenSampleSquare 2
    let jittersample = JitteredSampleSquare 2

    (*lights*)
    let intensity = 0.7

    let pointlight = mkPointLight (mkPoint 0.0 5.0 -5.0) white intensity

    let p1, p2, p3 = (mkPoint -2.5 5.0 0.0), (mkPoint 2.5 5.0 0.0), (mkPoint -2.5 5.0 -5.0)
    let randomlight = mkSquareLight randomsample white intensity p1 p2 p3
    let evenlylight = mkSquareLight evensample white intensity p1 p2 p3
    let jitterlight = mkSquareLight jittersample white intensity p1 p2 p3
    let arealight = jitterlight
    
    let ambient1 = mkAmbientLight white (1.0-intensity) //1.0-intensity so that adding the ambient light doesn't have a total intensity>1 (causes overflow)
    
    (*shapes*)
    let location1 = (mkPoint 0.0 0.5 1.0)
    let size1 = 1.0
    let sphere0 = mkSphere location1 size1
    let sphere1 = sphere0 (mkTexture (fun p -> red))
    let spheremarble = sphere0 (mkTexture marble)
    let sphereimgsandstone = sphere0 (mkTexture imgsandstone)

    let plane1 = mkPlane (mkPoint -5.0 -5.0 -5.0) (Vector.normalise (mkVector 0.0 1.0 0.0)) (mkTexture (fun p -> white))
    
    (*camera*)
    let camera1 = mkCamera (mkPoint 0.0 0.0 -2.0) (mkPoint 0.0 0.0 0.0) (mkVector 0.0 1.0 0.0) 2.0 3.0 3.0 500 500

    (*choose what to put in the scene*)
    let shapes = [|sphere1; plane1|]
    let lights = [|arealight|]

    let specularmax = 2.0 //the intensity of the specular highlight. 0.0 is no specular highlight, 1.0 is the natural amount, 2.0 is double intensity.
    let gloss = 10.0 //the sharpness of the specular highlight; a higher value means the specular rays point closer to the reflection direction.
    let specularsamplefun () = Scatter.mkJittered 1 //() to make it a function so evaluation is deferred
    let mirrorspecularsamplefun () = Scatter.mkEvenly 0

    let ambientlight = ambient1
    let camera = camera1

    let scene = mkScene shapes lights ambientlight camera (specularmax, gloss, specularsamplefun)

    //uncomment at least one of the following (commented out with //):


    (*render scene to window*)
    //renderToWindow scene


    (*alternatively, rendering scenes to files:*)
    let sceneswithstrings = [
        ("noise", mkScene [|(sphere0 (mkTexture noise)); plane1|] [|pointlight|] ambient1 camera (0.0, 0.0, specularsamplefun))
        ("fBm", mkScene [|(sphere0 (mkTexture fbm)); plane1|] [|pointlight|] ambient1 camera1 (0.0, 0.0, specularsamplefun))
        ("marble", mkScene [|spheremarble; plane1|] [|pointlight|] ambient1 camera (0.0, 0.0, specularsamplefun))
        ("sandstone", mkScene [|sphereimgsandstone; plane1|] [|pointlight|] ambient1 camera (0.0, 0.0, specularsamplefun))
        //
        ("gloss0", mkScene [|sphere1; plane1|] [|arealight|] ambient1 camera (1.0, 0.0, specularsamplefun))
        ("gloss1", mkScene [|sphere1; plane1|] [|arealight|] ambient1 camera (1.0, 1.0, specularsamplefun))
        ("gloss10", mkScene [|sphere1; plane1|] [|arealight|] ambient1 camera (1.0, 10.0, specularsamplefun))
        ("gloss100", mkScene [|sphere1; plane1|] [|arealight|] ambient1 camera (1.0, 100.0, specularsamplefun))
        ("gloss0vecreflection2", mkScene [|sphere1; plane1|] [|arealight|] ambient1 camera (1.0, 0.0, mirrorspecularsamplefun))
        //
        ("random", mkScene [|sphere1; plane1|] [|randomlight|] ambient1 camera (0.0, 0.0, specularsamplefun))
        ("evenly", mkScene [|sphere1; plane1|] [|evenlylight|] ambient1 camera (0.0, 0.0, specularsamplefun))
        ("jitter", mkScene [|sphere1; plane1|] [|jitterlight|] ambient1 camera (0.0, 0.0, specularsamplefun))   
    ] 
    //renderToFiles sceneswithstrings


    (*or measure speed*)
    let timerfuns = [
                ("random", fun () -> (renderScene (mkScene [|sphere1; plane1|] [|randomlight|] ambient1 camera (0.0, 0.0, specularsamplefun))))
                ("evenly", fun () -> (renderScene (mkScene [|sphere1; plane1|] [|evenlylight|] ambient1 camera (0.0, 0.0, specularsamplefun))))
                ("jitter", fun () -> (renderScene (mkScene [|sphere1; plane1|] [|jitterlight|] ambient1 camera (0.0, 0.0, specularsamplefun))))
                ]
    //measureSpeed timerfuns 3


    (*or render a visualization of sampling:*)
    let sample = mkEvenly 2
    //visualizeSample sample placementSamples


    (*visualization of a sample mapped to a hemisphere*)
    let sample2 = mkEvenly 0
    let spread = 0.0
    let upvec = mkVector 0.0 1.0 0.0
    let hemispherevecs = toHemisphereVecs sample2 upvec spread
    //visualizeSample (vecstoXZ hemispherevecs) placementPoints