﻿namespace Tracer

module Scene =
    open System.Windows.Forms
    open System.Drawing
    open Texture
    open Point
    open Scatter

    type Shape = Shape.IShape
    type Light = Lights.ILight
    type AmbientLight = Lights.AmbientLight
    type Camera = Camera.Camera

    type Scene = | SC of Shape array * Light array * AmbientLight * Camera * (float * float * (unit->Sample))

    let mkScene shapes lights ambientlight camera (specularMax, gloss, sample) =
        SC(shapes, lights, ambientlight, camera, (specularMax, gloss, sample))

    let getShapes (SC(sl,_,_,_,(_,_,_))) = sl
    let getLights (SC(_,ll,_,_,(_,_,_))) = ll
    let getAmbientlight (SC(_,_,al,_,(_,_,_))) = al 
    let getCamera (SC(_,_,_,c,(_,_,_))) = c
    let getSpecularMax (SC(_,_,_,_,(sm,_,_))) = sm
    let getGloss (SC(_,_,_,_,(_,g,_))) = g
    let getSamplefun (SC(_,_,_,_,(_,_,sfun))) = sfun

module Render =
    type Vector = Vector.Vector
    type Point = Point.Point
    type Ray = Ray.Ray
    type Color = Color.Color
    type Light = Lights.ILight
    type Shape = Shape.IShape
    type Scene = Scene.Scene
    type Texture = Texture.Texture
   
    let renderScene (scene:Scene) : System.Drawing.Bitmap = 

        let camera = Scene.getCamera scene
        let shapes = Array.toList (Scene.getShapes scene)
        let lights = Scene.getLights scene
        let ambientLights = Scene.getAmbientlight scene
        let specularmax = Scene.getSpecularMax scene
        let gloss = Scene.getGloss scene 
        let specularsamplefun = Scene.getSamplefun scene

        let imgheight, imgwidth = 500, 500
        let bitmap = new System.Drawing.Bitmap(imgheight, imgwidth)

        let bgcolor = Color.mkColor 100.0 100.0 100.0

        //taking the ray list [][] and mapping (using shapesCheck) each (ray list) to become the list of (closest hit for each ray)
        let hitsArrayArray (rays : Ray list[][]) : (Option<Shape*(float*Vector)>) list [][] = //2d array containing array of hits
            let shapesCheck (rays1 : Ray list) : Option<Shape*(float*Vector)> list = //we're going to use this for for each ray list in the [][]
                let gethitinfo (s:Shape) r = 
                    match s.Hit r with
                    | Some x -> Some(s,x)
                    | None -> None
                let rayhitshapes r = List.choose (fun (s:Shape) -> gethitinfo s r) shapes //one ray's hits on all shapes
                let hits = List.map( fun r -> rayhitshapes r) rays1 //all rays map via above
                let sort x = if List.isEmpty x then None else
                                let sorted = List.sortWith (fun (_,sh1) (_,sh2) -> if fst sh1 > fst sh2 then 1 else -1) x
                                Some sorted.Head
                let onlyclosests = List.map (fun x  -> sort x) hits //for each ray's hits on shapes we take its closest
                onlyclosests //should be the same length as ray list, and therefore we still know which hit matches which ray.
            let getHits (ra : Ray list[]) : Option<Shape*(float*Vector)> list[] = Array.map (fun x -> shapesCheck x) ra
            let transformedrows = Array.map (fun x -> getHits x) rays
            transformedrows

        //determine color for a hit
        let determineColor (ray:Ray) (s:Shape) (hit:float*Vector)=
            (*calculate things which will be used: distance, reflection vector, color, hitposition.*)
            let distance, n = fst hit, snd hit
            let reflectionVec =
                let unitn = Vector.normalise n
                let r = Ray.getDirection ray
                -2.0*(Vector.dotProduct r unitn)*unitn + r //
            let sourceColor = 
                let hitpoint = (Point.move (Ray.getOrigin ray) (Vector.multScalar (Ray.getDirection ray) distance))  
                let color = Texture.getColor s.Texture hitpoint
                color
            let hitPosition =
                let getHitPositionOn c d n ray : Point.Point=
                    let rayHitPosition = Point.move (Camera.getPosition c) (Vector.multScalar (Ray.getDirection ray) d)                 //Ray hit position on shape
                    let adjusted = Point.move rayHitPosition (Vector.multScalar n 0.0001)   //Move point along normal, so that it doesn't cast shadowray in shape 
                    adjusted
                getHitPositionOn camera distance n ray

            (*get different kinds of lighting*)
            let onelight (l:Light) =
                let shadowrays = l.getShadowRays hitPosition
                l.getLighting [Lighting.cosweight] shadowrays shapes hit 
            let onespecular (l:Light) =
                let reflectrays = Scatter.vecsToRaysFrom hitPosition (specularsamplefun ()) reflectionVec gloss
                //can't just use the reflectrays directly- shadowrays that arent blocked by a shape are assumed to touch a light. as such:
                let shadowreflectrays = List.filter(fun r -> l.getDoesHit r) reflectrays
                let proportionhit = ((float)shadowreflectrays.Length) / ((float)reflectrays.Length)
                let color = l.getLighting [(fun sr h -> proportionhit)] shadowreflectrays shapes hit
                color

            let getiflights (lightsiflit : Option<Color> list)=
                let exists o = match o with 
                                |Some o -> true
                                |None -> false
                match List.tryFind exists lightsiflit with
                    | Some l -> 
                        let getifsome a = List.choose (fun x->x) a
                        let lightingcolor = Color.addColors (getifsome lightsiflit)
                        lightingcolor
                    | None -> Color.mkColor 0.0 0.0 0.0

            let lightings = getiflights (List.map (fun l->onelight l) (Array.toList lights))
            let speculars = lazy(getiflights (List.map (fun l->onespecular l) (Array.toList lights)))
            
            (*apply lighting*)
            let basiclighting = //source color, ambient light, and normal lighting 
                let appliedAmbient = Lights.applyAmbient ambientLights lightings
                let ar, ag, ab = Color.getRGB (Color.scaleTo1 appliedAmbient)
                let sr, sg, sb = Color.getRGB sourceColor
                let ncr, ncg, ncb = ar * sr, ag * sg, ab * sb //multiplication: if it's being hit by a maximum intensity white, it'll end up the same as the source color, if something else it'll change
                Color.mkColor ncr ncg ncb

            let appliedspecularto acolor =
                if specularmax>0.0 then 
                    let specular = speculars
                    let adjustedspec = Color.scaleColor (specular.Force()) specularmax
                    Color.capColor (Color.addColor adjustedspec acolor)
                else acolor

            appliedspecularto basiclighting
         
        let colorgrid = //uses hitsArrayArray and determineColor
            let rays0 = Camera.getRays camera
            let rays = Array.map (fun x -> Array.map (fun y -> [y]) x ) rays0 //turns rays into ray lists

            let hits = hitsArrayArray rays
            let rowstocolor rayrow hitrow =
                let getColor r h = match h with
                                   | None -> bgcolor
                                   | Some x -> determineColor r (fst x) (snd x)
                let listtocolor rlist hlist = let allColors = List.map2 (fun r h -> getColor r h) rlist hlist //for each cell, map rays/hit pairs to colors
                                              let avgColor = Color.avgFromList allColors
                                              Color.toSysColor avgColor
                Array.map2 (fun raylist hitlist -> listtocolor raylist hitlist) rayrow hitrow
            Array.map2 (fun x y -> rowstocolor x y) rays hits
            
        let colorize = //colors the bitmap using colorgrid
            let rowscolorize rownumber row =
                let cpxlheight, cpxlwidth = Camera.getPixelHeight camera, Camera.getPixelWidth camera
                let regionheight, regionwidth = imgheight/cpxlheight , imgwidth/cpxlwidth //bitmap pixels per viewpixel
                let patchcolorize row column color = 
                    let regionoriginy, regionoriginx = column*regionheight, row*regionwidth
                    let regionendy, regionendx = regionoriginy + regionheight, regionoriginx + regionwidth
                    for x in regionoriginx..regionendx-1 do
                        for y in regionoriginy..regionendy-1 do
                        bitmap.SetPixel(x, y, color)
                Array.iteri (fun j y -> patchcolorize rownumber j y) row
            Array.iteri (fun i x -> rowscolorize i x) colorgrid

        colorize
        bitmap

    let renderToWindow (scene: Scene) =
        let bitmap = renderScene scene

        let camera = Scene.getCamera scene
        let width = Camera.getPixelWidth camera
        let height = Camera.getPixelHeight camera

        let form = 
            new System.Windows.Forms.Form(
                ClientSize = new System.Drawing.Size(width, height),
                FormBorderStyle = System.Windows.Forms.FormBorderStyle.FixedSingle,
                Text = "Rendered image")

        let paint (e: System.Windows.Forms.PaintEventArgs) = e.Graphics.DrawImage(bitmap, 0, 0)
        form.Paint.Add(paint)
        System.Windows.Forms.Application.Run form

    let renderToFile ((filename : string), (scene : Scene))  = 
        let bitmap = renderScene scene;
        bitmap.Save(filename+".png")

    let renderToFiles allscenesandstring = List.iter (fun x -> renderToFile x) allscenesandstring

    let measureSpeed timerfuns iterations=
        let manytimerfuns n = 
            let makemore n tf = List.init n (fun index -> ((fst tf+(sprintf " %i" (index+1))), snd tf))
            List.concat(List.map (fun tf -> makemore n tf) timerfuns)
        let log strfuntup =
            let stopWatch = System.Diagnostics.Stopwatch.StartNew()
            ignore ((snd strfuntup) ())
            stopWatch.Stop()
            sprintf "%s: %f\n" (fst strfuntup) stopWatch.Elapsed.TotalMilliseconds
        let text = List.fold (fun acc sf -> acc + log sf) "" (manytimerfuns iterations)
        System.IO.File.Delete("speed.txt")
        System.IO.File.AppendAllText("speed.txt", text) 

