﻿namespace Tracer

module Scatter =

    type Point = Point.Point
    type Sample = (float*float) list //points in the unit square

    let rnd = System.Random() //not purely functional programming here

    (*these take a space and put a sample in it.*)
    let placementRandom (a:float, b:float) heightwidth : float*float= //they take (a,b) instead of (x1,y1) (x2,y2) because we assume the space is a square, and so knowing the least corner and the height(=width) is enough to know the location of all the corners.
        let hw = heightwidth
        let granularity = 1000
        let sizepergrain = hw/(float)granularity
        let randomoffsetx = (float)(rnd.Next(0,granularity))
        let randomoffsety = (float)(rnd.Next(0,granularity))
        let xoffset, yoffset = randomoffsetx*sizepergrain, randomoffsety*sizepergrain  //note: inclusive lower bound, exclusive upper bound.
        (a+xoffset, b+yoffset)
    let placementCenter (a:float, b:float) heightwidth : float*float= //a,b is the bottom left coord of a square
        (a+heightwidth, b+heightwidth)

    (*divide a space.*)
    let divisions (a:float, b:float) num : float * ((float*float) list) = //(a,b) for the unit square, for instance, is (0,1)
        let cnum = pown 2 num //number of "checkerboard square" across or high (interchangeable bc square)
        let eachwidth = (b-a)*(1.0/(float)cnum)  //they're all the same width (and height)
        let coordlist = 
            let tiles =
                let rec getTiles n c=
                    seq { for i in 1 .. c do yield (n, i) }
                let s = seq { 1 .. cnum }
                let tiles0 = Seq.map (fun n -> getTiles n cnum ) s
                Seq.concat tiles0
            let tileToCoord (a,b) =  
                let bottomLeftOfTile = (((float)a-1.0)*eachwidth), (((float)b-1.0)*eachwidth)
                bottomLeftOfTile
            Seq.map (fun (a,b) -> tileToCoord (a,b) ) tiles

        ((float) eachwidth, Seq.toList coordlist)

    (*make a sample*)
    let mkRandom (n:int) : Sample = [for i in 1 .. n -> placementRandom (0.0, 0.0) 1.0 ]
    let mkEvenly (splits:int) : Sample = 
        let coordlist = (divisions (0.0, 1.0) splits)
        List.map (fun c -> placementCenter c (fst coordlist)) (snd coordlist)
    let mkJittered (splits:int) : Sample = 
        let coordlist = (divisions (0.0, 1.0) splits)
        List.map (fun c -> placementRandom c (fst coordlist)) (snd coordlist)

    (*place a sample in space*)
    let toPoints (s:Sample) (start:Point) (x1:Point) (y1:Point) : Point list = //points are origin, bottom right of grid, top left of grid
        let sx, sy, sz = Point.getX start, Point.getY start, Point.getZ start
        let xx, xy, xz = Point.getX x1, Point.getY x1, Point.getZ x1
        let yx, yy, yz = Point.getX y1, Point.getY y1, Point.getZ y1
        let xscale = Point.distance start x1
        let yscale =  Point.distance start y1
        let scalelocation a b =
            let movedbyxv = Point.move start (Vector.multScalar xscale a) //move along x vector in proportion to a
            let movedbyyv = Point.move movedbyxv (Vector.multScalar yscale b)
            movedbyyv
        List.map (fun (a,b) -> (scalelocation a b)) s

    (*make a sample with a location in space*)
    let RandomSampleSquare n startp x1 y1 = 
        toPoints (mkRandom n) startp x1 y1 
    let JitteredSampleSquare s startp x1 y1  =
        toPoints (mkJittered s) startp x1 y1 
    let EvenSampleSquare s startp x1 y1  =
        toPoints (mkEvenly s) startp x1 y1 

    (*maps a sample to a hemisphere of vectors*)
    let toHemisphereVecs (s:Sample) (normal:Vector.Vector) (spread:float) : Vector.Vector list =
        let toVec x y spread=
            let e = spread //spread 0 is a uniform distribution (not bunched up closer to the top of the hemisphere).
            let azi = 2.0*System.Math.PI*x
            let cospol = y**(1.0/(e+1.0)) //instead of acos and then cos on this, just use it directly and find sin the "hard" way
            let sinpol = sqrt(1.0-cospol) //by pythogorean identity

            //printfn "x y %f %f azi pol %f %f\n" x y azi cospol
            let x = sinpol * (cos azi)
            let y = cospol
            let z = sinpol * (sin azi)

            let w = Vector.normalise normal
            let v = 
                let arbitraryvec = Vector.mkVector ((Vector.getX w)+10.0) (Vector.getY w)(Vector.getZ w)
                Vector.crossProduct w (Vector.normalise arbitraryvec)
            let u = Vector.crossProduct w v

            let yw, xv, zu = (y*w), (x*v), (z*u)
            let vec = yw+xv+zu
            vec
        List.map (fun (x, y)->toVec x y spread) s

    //gets the hemisphere vectors and converts them to rays.
    let vecsToRaysFrom (origin:Point) (s:Sample) (normal:Vector.Vector) (spread:float)= 
        let vl = toHemisphereVecs s normal spread
        let newp = Point.move origin (Vector.normaliseto normal 0.0001)
        List.map (fun v -> Ray.mkRay newp v) vl


    (*The rest of the module is for testing samples by painting a representation of them onto an image.*)


    //for reducing vectors to 2D endpoints:
    let vecEndPoint vecs = List.map (fun v -> Point.move (Point.mkPoint 0.0 0.0 0.0) v) vecs //endpoints
    let vecstoXY vecs = List.map (fun p -> Point.getX p, Point.getY p) (vecEndPoint vecs) //find some dimensions of points
    let vecstoXZ vecs = List.map (fun p -> Point.getX p, Point.getZ p) (vecEndPoint vecs)
    let vecstoZY vecs = List.map (fun p -> Point.getZ p, Point.getY p) (vecEndPoint vecs)

    //mapping samples to a canvas size:
    let placementSamples size x y = (int)((float)size*x), (int)((float)size*y) //size refers to canvas size. this one is used for sample points in the range 0 to 1
    let placementPoints size x y = //this one is used for the converted hemisphere vectors, since those can range from -1 to 1
        let half = (float)(size/2)
        (int)(half+x*half), (int)(half+y*half)

    //paint points to canvas, given a function for placing points on a canvas:
    let visualizeSample s placementfun= //remember, higher numbers are further DOWN in a bitmap, so as of right now, it's flipped, but this doesn't actually matter.
        let size = 1001
        let width = size
        let height = size
        let bitmap = new System.Drawing.Bitmap(height, width)
        let bgcolor = System.Drawing.Color.Black
        let color = System.Drawing.Color.White

        let bitmap2 =
            for x in {0..width-1} do 
                for y in {0..height-1} do
                    bitmap.SetPixel(x, y, bgcolor)

            for (x,y) in s do 
                let a,b = placementfun size x y
                bitmap.SetPixel(a, b, color)
            bitmap

        let form = 
            new System.Windows.Forms.Form(
                ClientSize = new System.Drawing.Size(width, height),
                FormBorderStyle = System.Windows.Forms.FormBorderStyle.FixedSingle,
                Text = "Rendered image")

        let paint (e: System.Windows.Forms.PaintEventArgs) = e.Graphics.DrawImage(bitmap2, 0, 0)
        form.Paint.Add(paint)
        System.Windows.Forms.Application.Run form