﻿namespace Tracer

module Shape = 
    type Vector = Vector.Vector
    type Point = Point.Point
    type Ray = Ray.Ray
    type Texture = Texture.Texture

    type IShape =
        abstract member Hit: (Ray -> Option<float*Vector>)
        abstract member Texture: (Texture)

    let mkSphere (center : Point) (radius : float) (tex : Texture) =
        let sphereHit (ray:Ray) =
            let v = Point.distance center (Ray.getOrigin ray) 
            let d = Vector.normalise (Ray.getDirection ray)
            let rSquare = System.Math.Pow(radius, 2.0)
            let vSquare = Vector.dotProduct v v

            //Check if discriminant is non-negative
            let disc = System.Math.Sqrt( (System.Math.Pow((Vector.dotProduct v d), 2.0)) - (vSquare - rSquare) )
            if disc < 0.0 then 
                None
            else
                //Calc t1 & t2
                let t1 = -(Vector.dotProduct v d) - System.Math.Sqrt( (System.Math.Pow((Vector.dotProduct v d), 2.0)) - (vSquare - rSquare) )
                let t2 = -(Vector.dotProduct v d) + System.Math.Sqrt( (System.Math.Pow((Vector.dotProduct v d), 2.0)) - (vSquare - rSquare) )

                //Check for the smallest non-negative t
                if t1 > 0.0 && (t1 < t2 || t2 < 0.0) then
                    let y1 = Point.move (Ray.getOrigin ray) (Vector.multScalar d t1)
                    let n1 = Vector.normalise(Point.distance center y1)
                    Some(t1,n1) 
                elif (t2 > 0.0 && (t2 < t1 || t1 < 0.0)) then
                    let y2 = Point.move (Ray.getOrigin ray) (Vector.multScalar d t1)
                    let n2 = Vector.normalise(Point.distance center y2)
                    Some(t2,n2)
                else 
                    None

        {
         new IShape with 
         member this.Hit = sphereHit
         member this.Texture = tex
         }

    let mkPlane (p:Point) (n:Vector) (tex:Texture) : IShape =
        let planehit (r:Ray) : Option<float*Vector> =
            let l0 = Ray.getOrigin r
            let l = Ray.getDirection r
            let disPL = Point.distance l0 p
            let num = Vector.dotProduct disPL n
            let denom = Vector.dotProduct l n
            let d = num/denom

            if d < System.Double.Epsilon  then None else Some(d, n)
        {
         new IShape with 
         member this.Hit = planehit
         member this.Texture = tex
         }


    let hitSide (a : Point) (b : Point) (c : Point) (r:Ray.Ray) = //used by the square area light (see Light.fs)
            // Side vectors AB AC
            let v1 = Point.distance a b
            let v2 = Point.distance a c
            
            // Ray 
            let rd = Ray.getDirection r
            let ro = Ray.getOrigin r

            let n = Vector.crossProduct v1 v2  // Plane n-vector
            let nd = Vector.dotProduct n rd    //

            // If ray direction follows plane / is parallel with the plane / no intersection between ray and plane
            if(System.Math.Abs(nd) < System.Double.Epsilon) then None else

            let t = (Vector.dotProduct n (Point.distance ro a)) / nd // Distance
            let p = Point.move ro (Vector.multScalar rd t)           // Intersection point

            // Intersection point behind camera
            if(t < 0.0) then None else
            let u = Vector.dotProduct (Point.distance a p) v1
            let v = Vector.dotProduct (Point.distance a p) v2

            if((0.0 <= u && u <= Vector.dotProduct v1 v1) && (0.0 <= v && v <= Vector.dotProduct v2 v2)) then
                Some(t, Vector.normalise n) else
                None 