﻿namespace Tracer

module Texture =

    open Color
    open Point

    type Texture = T of (Point -> Color)

    let mkTexture f = T(f)
    let getColor (T(f)) point = f point

    let seed = 1
    let rnd = System.Random(seed)

    (*lattice points*)
    let permutationtable = 
        [|
        151;160;137;91;90;15;  
        131;13;201;95;96;53;194;233;7;225;140;36;103;30;69;142;8;99;37;240;21;10;23;    // Random numbers from 0-255 inclusive; as defined by Ken Perlin...
        190; 6;148;247;120;234;75;0;26;197;62;94;252;219;203;117;35;11;32;57;177;33;    // Could generate one from scratch; but there's no real point in doing so.
        88;237;149;56;87;174;20;125;136;171;168; 68;175;74;165;71;134;139;48;27;166;
        77;146;158;231;83;111;229;122;60;211;133;230;220;105;92;41;55;46;245;40;244;    //the reason it's an array and not a list is, ofc, b/c arrays are better for lookup.
        102;143;54; 65;25;63;161; 1;216;80;73;209;76;132;187;208; 89;18;169;200;196;
        135;130;116;188;159;86;164;100;109;198;173;186;3;64;52;217;226;250;124;123;
        5;202;38;147;118;126;255;82;85;212;207;206;59;227;47;16;58;17;182;189;28;42;
        223;183;170;213;119;248;152; 2;44;154;163; 70;221;153;101;155;167; 43;172;9;
        129;22;39;253; 19;98;108;110;79;113;224;232;178;185; 112;104;218;246;97;228;
        251;34;242;193;238;210;144;12;191;179;162;241; 81;51;145;235;249;14;239;107;
        49;192;214; 31;181;199;106;157;184; 84;204;176;115;121;50;45;127; 4;150;254;
        138;236;205;93;222;114;67;29;24;72;243;141;128;195;78;66;215;61;156;180
        |]
    
    (*init lattice point values*)
    let period = permutationtable.Length
    let valuetable =
        let randomvalue x = rnd.NextDouble()
        Array.init period (fun x -> randomvalue x)

    (*transform a point to lie within the period*)
    let transformedp pt =
        let zoom = 10.0 //i think zooming in (so that a sphere 1 unit across is, say, n values instead of 1 just means multiplying the world xyz by n)...
        let tf x = abs((x*zoom)%(float)period)
        let x, y, z = tf (Point.getX pt), tf (Point.getY pt), tf (Point.getZ pt)
        Point.mkPoint x y z 

    (*given a lattice point (with x y and z values within the period) find the value of the point*)
    let latticecoordvalue (x,y,z) = 
        let xval = permutationtable.[x]
        let yval = permutationtable.[(xval+y)%256]  //(modulo 256 is the same as bitwise and with 255)
        let zval = permutationtable.[(yval+z)%256]
        valuetable.[zval]

    (*find the corners of a cell for a point in the period*)
    let periodcoords (p:Point) =
        //for each z y z, figure out which ones should be rounded up or down (marked as true or false). "true true false" for instance, is later translated to being the right-top-back corner.
        let boolsets =  
            let dimensions = 3
            let permute n = 
                let funlist = List.init (pown 2 dimensions) (fun i -> [])
                let rec p n (flist:bool list list) = //n dimensions
                    match n with 
                    | 0 ->  flist
                    | _ ->  let length = flist.Length
                            let list1 = List.mapi (fun i x -> if i<length/2 then true::x else false::x) flist   //so the first half is true and the other half is false. 
                            let fsthalf = List.filter (fun (x::t) -> x) list1
                            let sndhalf = List.filter (fun (x::t) -> not x) list1
                            p (n-1) fsthalf @ p (n-1) sndhalf   //then we run the function again on each half. for instance, start w/ 4 items -> t t f f -> tt tf @ ft ff
                p n funlist
            permute dimensions
        //for each item in boolset make a point made of the point but with x y z floor or ceilinged depending on the bool.
        let makePoint boolset xyz =
            let round b x = if b then floor x else floor (x+1.0)
            let newxyz = List.map2 (fun b c -> (round b c)%256.0 ) boolset xyz //modulo 256 so rounding up doesnt push a point out of bounds.
            Point.mkPoint newxyz.[0] newxyz.[1] newxyz.[2]
        let xyzlist = [Point.getX p; Point.getY p; Point.getZ p]
        List.map (fun x -> makePoint x xyzlist) boolsets

    (*linear interpolation given the point and its 8 corners*)
    let linearinterpolate (corners: Point list) (p : Point) (getvalue:int*int*int->float) =
        let interpolate (l:(Point*float) list) f1 f2 f3 = //replace each [(point, value) (point,value)]  with a single (point,value).
            let interpolate1D (x1, x1val) (x2, x2val) proportion=
                let inverseproportion = 1.0-proportion
                let value = ((inverseproportion*x1val) + (proportion*x2val))
                let position = f3 p
                position, value
            let old1, old2 = l.[0], l.[1]
            let x, y = f1 (fst old1), f2 (fst old2)
            let less, more = f3 (fst old1), f3 (fst old2)
            let proportion = 1.0-((more-f3 p)/(more-less))
            if more = less then failwith "dividing by zero"
            let i = interpolate1D (less, snd old1) (more, snd old2) proportion
            (x, y, (fst i)), snd i
        
        let chunkfun = List.chunkBySize 2
        let toInt (a,b,c) = (int)a, (int)b, (int)c 
        let cornerswithvalues = List.map (fun x -> x, (getvalue (toInt(getCoord x))) ) corners

        let mkpointorder (a, b, c) str= 
            match str with
            | "XYZ" -> mkPoint a b c
            | "XZY" -> mkPoint a c b
            | "YZX" -> mkPoint c a b
            | _ -> failwith "invalid point order"

        let Zpolate chunk = 
            let floats, value = interpolate chunk getX getY getZ
            mkpointorder floats "XYZ", value    //because interpolate takes and returns things in the same order each time, it's called with x y and z in a different order in each time, and the result is then put back in the correct order afterwards.
        let Ypolate chunk = 
            let floats, value = interpolate chunk getX getZ getY
            mkpointorder floats "XZY", value
        let Xpolate chunk = 
            let floats, value = interpolate chunk getY getZ getX
            mkpointorder floats "YZX", value

        let Xpolated = List.map (fun chunk -> Xpolate chunk) (chunkfun cornerswithvalues) //8 to 4
        let XYpolated = List.map (fun chunk -> Ypolate chunk) (chunkfun Xpolated)         //4 to 2
        let XYZpolated = List.map (fun chunk -> Zpolate chunk) (chunkfun XYpolated)       //2 to 1
        
        //error checking
        let checkposition = ////ZYXpolated.Head should have the same coordinates as the original point p
            let x0, y0, z0 = getCoord p
            let x, y, z = getCoord (fst XYZpolated.Head)
            if (abs x0-x > 0.01) || (abs y0-y > 0.01) || (abs y0-y > 0.01) then failwith "points don't match"
        if XYZpolated.Length>1 then failwith "XYZpolated: too many remaining values" else checkposition
        
        snd XYZpolated.Head

    (*below are all the p->float texture functions.*)

    let valuenoise p = 
        let tp = transformedp p
        linearinterpolate (periodcoords tp) tp latticecoordvalue

    (*
    //made redundant by fBm
    let fractalsum octaves p =
        let rec frac noise i frequency amplitude =
            if i<octaves then noise 
                else
                    let freq = frequency * 2.0
                    let amp = amplitude * 0.5
                    let addnoise = amp*(makenoise (p*freq))
                    frac (noise+addnoise) (i+1) freq amp
        frac (makenoise p) 0 1.0 1.0
    *)

    let fBm octaves p lacunarity gain turbulence=
        //generalized version of fractal sum: to get same result, lacunarity should be 2 and gain 0.5.
        //also, has the option for turbulence to avoid making two seperate functions needlessly.      
        let rec frac noise noisefunc i frequency amplitude maxpossible =
            if i>=octaves then 
                noise/maxpossible
                else
                    let freq = frequency * lacunarity
                    let amp = amplitude * gain
                    let morenoise = amp*(noisefunc (p*freq))
                    let noisetoadd= if turbulence then abs(2.0*morenoise - 1.0) else morenoise    //turbulence: changes the range from 0-1 to -1-1, then take the absolute value (so it's still 0-1 in the end)
                    frac (noise+noisetoadd) noisefunc (i+1) freq amp (maxpossible+maxpossible*gain)
        frac (valuenoise p) valuenoise 1 1.0 1.0 1.0

    let stonetex p repeatsfreq noiseproportion=
        let x = getX p
        //instead of using the value returned from valuenoise as the color, we want to use it as a jitter on the x given to the sine-wave-pattern function.
        let noiseval = fBm 5 p 2.0 0.5 false
        //let noiseval = valuenoise p
        //let noiseval = rnd.NextDouble()
        let sinfit x = (1.0+x)/2.0
        let inversep = 1.0-noiseproportion        
        let newx = (x*inversep + noiseval*noiseproportion)*repeatsfreq // and *repeatsfreq to give sin(x) a shorter wavelength (oscillates "faster").
        sinfit (sin(newx))
    
    let marbletex (p:Point) = stonetex p 100.0 0.4
    let sandstonetex (p:Point) = stonetex p 0.5 0.1


    (*below, functions to convert floats to colors.*)
    let tomonochrome (n:float) = 
        let c = n*255.0
        mkColor c c c
    
    let defaultramp = 
        let black, white, red = mkColor 0.0 0.0 0.0, mkColor 255.0 255.0 255.0, mkColor 255.0 0.0 0.0
        [|(0.0, black); (0.7, red); (1.0, white)|]
        //[|(0.0, red); (1.0, white)|]

    let tocolor (n:float) (ramp: (float*Color) array) =
        let under, over = Array.partition (fun (f,c) -> if f<n then true else false) ramp
        //less and more should be the colors that are closest to n on either side (of the number line). for that to be true we assume the color ramp is already sorted .
        let lesser = 
            let len = Array.length under
            under.[len-1]
        let greater = over.[0]

        let gap = fst greater - fst lesser
        let filled = n - fst lesser
        let proportion = 1.0-(filled/gap)

        let lessercolor = scaleColor (snd lesser) proportion
        let greatercolor = scaleColor (snd greater) (1.0-proportion)
        addColor lessercolor greatercolor

    let toimgcolor (n:float) (img:System.Drawing.Bitmap) =
        let width = img.Width - 1
        let colorlocation = n*(float)width //n, which is between 0 and 1, says how far along to image to move to colorpick.
        let roundedlocation = (int)(System.Math.Round colorlocation)
        let color = img.GetPixel(roundedlocation, 1) //y is just 1 because we only care about 1 dimension.
        Color.fromSysColor color