﻿namespace Tracer

module Vector =
    type Vector = | V of float * float * float with
      override v.ToString() =
        match v with
          V(x,y,z) -> "["+x.ToString()+","+y.ToString()+","+z.ToString()+"]"

    let mkVector x y z = V(x, y, z)
    let getX (V(x,_,_)) = x
    let getY (V(_,y,_)) = y
    let getZ (V(_,_,z)) = z
    let getCoord (V(x, y, z))   = (x, y, z) 
    let multScalar (V(x,y,z)) s = V(x*s, y*s, z*s)
    let magnitude (V(x, y, z))  = sqrt(System.Math.Pow(x, 2.0)+System.Math.Pow(y, 2.0)+System.Math.Pow(z, 2.0))
    let dotProduct (V(x1, y1, z1)) (V(x2, y2, z2))      = (x1*x2) + (y1*y2) + (z1*z2)
    let cosBtwn v1 v2 = (dotProduct v1 v2) / (magnitude v1 *magnitude v2)
    let crossProduct (V(x1, y1, z1)) (V(x2, y2, z2))    = V(y1*z2-z1*y2, z1*x2-x1*z2, x1*y2-y1*x2)
    let normalise (V(x,y,z) as v) = let mv = magnitude v
                                    V(x/mv, y/mv, z/mv)
    let normaliseto (V(x,y,z) as v) s = let nv = normalise v
                                        V(x*s, y*s, z*s)
    let round (V(x,y,z)) (d:int) = V(System.Math.Round(x,d),System.Math.Round(y,d),System.Math.Round(z,d))

    type Vector with
      static member ( ~- ) (V(x,y,z)) = V(-x,-y,-z)
      static member ( + ) (V(ux,uy,uz),V(vx,vy,vz)) = V(ux+vx, uy+vy, uz+vz)
      static member ( - ) (V(ux,uy,uz),V(vx,vy,vz)) = V(ux-vx, uy-vy, uz-vz)
      static member ( * ) (s, v) = multScalar v s
      static member ( * ) (u, v) = dotProduct u v


module Point =
    type Vector = Vector.Vector

    type Point =
      | P of float * float * float
      override p.ToString() =
        match p with
          P(x,y,z) -> "("+x.ToString()+","+y.ToString()+","+z.ToString()+")"

    let mkPoint x y z = P(x,y,z)
    let getX (P(x,_,_)) = x
    let getY (P(_,y,_)) = y
    let getZ (P(_,_,z)) = z
    let getCoord (P(x,y,z)) = (x,y,z)
    let move (P(x,y,z)) V = P(x+(Vector.getX V), y+(Vector.getY V), z+(Vector.getZ V))
    let moveByValue (P(x,y,z)) v = P(x+v, y+v, z+v)
    let distance (P(px,py,pz)) (P(qx,qy,qz)) = Vector.mkVector (qx-px) (qy-py) (qz-pz)
    let direction p q = Vector.normalise(distance p q)
    let round (P(px,py,pz)) (d:int) = P(System.Math.Round(px, d), System.Math.Round(py, d), System.Math.Round(pz, d))

    let min (p1: Point) (p2: Point) =
        let x1, y1, z1 = getCoord p1
        let x2, y2, z2 = getCoord p2
        Point.P(min x1 x2, min y1 y2, min z1 z2)

    let max (p1: Point) (p2: Point) =
        let x1, y1, z1 = getCoord p1
        let x2, y2, z2 = getCoord p2
        Point.P(max x1 x2, max y1 y2, max z1 z2)

    type Point with
      static member ( + ) (P(ux,uy,uz), P(vx,vy,vz)) = P(ux+vx, uy+vy, uz+vz)
      static member ( - ) (P(ux,uy,uz), P(vx,vy,vz)) = P(ux-vx, uy-vy, uz-vz)
      static member ( * ) (P(ux,uy,uz), n) = P(ux*n, uy*n, uz*n)
      

module Ray =
    type Vector = Vector.Vector
    type Point = Point.Point

    type Ray = 
        | R of Point * Vector * Vector
        | S of Point * Vector * Vector * float

    let mkRay p v = 
        R(p, v, Vector.mkVector (1.0 / Vector.getX v) (1.0 / Vector.getY v) (1.0 / Vector.getZ v))

    let mkShadow p v i d = S(p, v, i, d)

    let getOrigin r = 
        match r with
        | R(p, v, i)    -> p
        | S(p, v, i, d) -> p

    let getDirection r = 
        match r with
        | R(p, v, i)    -> v
        | S(p, v, i, d) -> v

    let getInverted r =
        match r with
        | R(p, v, i)    -> i
        | S(p, v, i, d) -> i

    let getDistance r = 
        match r with
        | R(p, v, i)    -> 0.0
        | S(p, v, i, d) -> d
